/*
 *  Video stacking program with change detection.
 *
 *  Panu Lahtinen, pnuu+finsprite@iki.fi
 */

#include "finsprite.h"

void init_peak_hold(void){
  peak_hold->img_buffer = calloc(current_frame.width*current_frame.height,sizeof(*(peak_hold->img_buffer)));
  peak_hold->width = current_frame.width;
  peak_hold->height = current_frame.height;
  peak_hold->timestamp = current_frame.timestamp;
  peak_hold->usec = current_frame.usec;
  stack_timestamp = current_frame.timestamp;
  stack_usec = current_frame.usec;
}

void init_min_hold(void){
  int i = 0;
  min_hold->img_buffer = calloc(current_frame.width*current_frame.height,sizeof(*(min_hold->img_buffer)));
  for(i=0;i<current_frame.width*current_frame.height;i++){
    min_hold->img_buffer[i] = 255;
  }
  min_hold->width = current_frame.width;
  min_hold->height = current_frame.height;
  min_hold->timestamp = current_frame.timestamp;
  min_hold->usec = current_frame.usec;
  stack_timestamp = current_frame.timestamp;
  stack_usec = current_frame.usec;
}

void init_latest_img(void){
  latest_img->img_buffer = calloc(current_frame.width*current_frame.height,sizeof(*(latest_img->img_buffer)));
  latest_img->width = current_frame.width;
  latest_img->height = current_frame.height;
  latest_img->timestamp = current_frame.timestamp;
  latest_img->usec = current_frame.usec;
  latest_img_timestamp = current_frame.timestamp;
  latest_img_usec = current_frame.usec;
}

void create_md_zones(int img_width, int img_height){

  unsigned char *mask = NULL;
  int mask_width = 0;
  int mask_height = 0;

  int i = 0;

  /* Read mask, if available */
  if(md_mask_fname != NULL) {
    /* read */
    mask = read_mask(md_mask_fname, &mask_width, &mask_height);
    printf("Mask read from %s (%dx%d)\n",md_mask_fname,mask_width,mask_height);
  }

  if(mask_width != img_width || mask_height != img_height){
    if(md_mask_fname != NULL){
      printf("Mask is not the same size as image, deleting mask\n");
    }
    if(mask != NULL){ free(mask); }
    /* Default to allow all pixels to be used in motion detection */
    mask = (unsigned char*)malloc(img_width*img_height*sizeof(unsigned char));
    for(i=0;i<img_width*img_height;i++){ mask[i] = 1; }
    mask_width = img_width;
    mask_height = img_height;
  }

  /* Setup zone sums */
  md_zone_sums = (int*)malloc(md_num_x_zones*md_num_y_zones*sizeof(int));
  md_zone_sums_new = (int*)malloc(md_num_x_zones*md_num_y_zones*sizeof(int));
  for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
    md_zone_sums[i] = INT_MAX;
    md_zone_sums_new[i] = 0;
  }

  /* Number of valid pixels in each motion detection zones */
  md_zone_num_pix = (unsigned int*)malloc(md_num_x_zones*md_num_y_zones*sizeof(unsigned int));
  for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
    md_zone_num_pix[i] = 0;
  }

  /* Setup zone reference */
  md_zone_reference = (int*)malloc(img_width*img_height*sizeof(int));

  /* Pixels per zone in each direction */
  float pix_per_x_zone = (img_width/(float)md_num_x_zones);
  float pix_per_y_zone = (img_height/(float)md_num_y_zones);

  int x_zone = 0;
  int y_zone = 0;
  int row = -1;

  /* Form the reference vector that shows to which zone each pixel belongs to */
  for(i=0; i < img_width*img_height; i++){
    if(i%img_width == 0){
      row++;
    }

    if(mask[i] != 0){
      x_zone = (int)((i-row*img_width)/pix_per_x_zone);
      y_zone = (int)(row/pix_per_y_zone);
      md_zone_reference[i] = x_zone + y_zone * md_num_x_zones;
      md_zone_num_pix[md_zone_reference[i]]++;
    } else {
      md_zone_reference[i] = -1;
    }
  }
  if(mask != NULL){
    free(mask);
  }
  /*
  for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
    printf("i: %d, num: %d\n",i,md_zone_num_pix[i]);
  }
  */

  printf("Zones ready\n");
}

void *md_process(void* ptr){

  int i = 0;
  double md_trigger_value = -1.0;

  char *tmp = NULL;

  md_process_on = 1;

  if(use_trig){

    /* Check if the current daily motion detection counter is up-to-date */
    if(md_trig_number == 0){
      //tmp = (char*)malloc(FNAME_SIZE*sizeof(char));
      char tmp3[FNAME_SIZE] = {'\0'};
      make_fname(tmp3,current_frame.timestamp,current_frame.usec,DAY_DIR);
      md_trig_number = find_md_folder_num(tmp3);
      //printf("md_trig_number: %d\n",md_trig_number);
      //free(tmp); tmp = NULL;
    }

    /*
    if(md_frames[md_i].img_buffer != NULL){
      free(md_frames[md_i].img_buffer);
      md_frames[md_i].img_buffer = NULL;
    }
    */
    if(md_frames[md_i].img_buffer == NULL){
      md_frames[md_i].img_buffer = calloc(current_frame.width*current_frame.height,sizeof(*md_frames[md_i].img_buffer));
      md_frames[md_i].width = current_frame.width;
      md_frames[md_i].height = current_frame.height;
    }

    /* Form the motion detection zones */
    if(md_zone_reference == NULL){
      create_md_zones(current_frame.width, current_frame.height);
    } else {
      /* Or if already done, update comparison values and reset for new image
      */
      for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
	md_zone_sums[i] = md_zone_sums_new[i];
	md_zone_sums_new[i] = 0;
      }
    }

    /* Update the timestamp */
    md_frames[md_i].timestamp = current_frame.timestamp;
    md_frames[md_i].usec = current_frame.usec;
  }

  /* Copy current image buffer to ring memory */
  for(i=0; i<current_frame.width*current_frame.height; i++){
    md_frames[md_i].img_buffer[i] = current_frame.img_buffer[i];
    /* If there are no detection at the moment, update zone sums */
    if(md_detection == -1){
      if(md_zone_reference[i] > 0){
	md_zone_sums_new[md_zone_reference[i]] += current_frame.img_buffer[i];
      }
    }
  }

  /* Detect rapid changes in illumination (motion detection) */
  if(md_buffer_ready){
    for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
      /* Use only zones with more than 25 % of pixels in non-masked zone */
      if(md_zone_num_pix[i] > 0.25*
	 current_frame.width*current_frame.width/(md_num_x_zones*md_num_y_zones) ){ // > 0
	/* Trigger if the new image zone was brighter */
	md_trigger_value = 100.0*md_zone_sums_new[i]/md_zone_num_pix[i] - 
	  100.0*md_zone_sums[i]/md_zone_num_pix[i];
	if(md_trigger_value > md_threshold){
	  
	  tmp = make_time_string(current_frame.timestamp,current_frame.usec,TRIG);
	  /* Get the date for the frame that triggers the saving. All the 
	     frames from this event are saved to same folder, no matter what 
	     date. */
	  sscanf(tmp,"%d-%d-%d",&md_trig_year,&md_trig_month,&md_trig_day);

	  md_triggered_zone = i;
	  printf("Zone %d triggered: %f (%d) at %s\n",i,100.0*md_zone_sums_new[i]/md_zone_num_pix[i]-100.0*md_zone_sums[i]/md_zone_num_pix[i],md_threshold,tmp);
	  free(tmp); tmp = NULL;

	  if(save_log){
	    if(md_log_fname == NULL){
	      md_log_fname = (char*)malloc(FNAME_SIZE*sizeof(char));
	      memset(md_log_fname,'\0',FNAME_SIZE);
	    }

	    char tmp2[FNAME_SIZE] = {'\0'};

	    make_fname(tmp2,current_frame.timestamp,current_frame.usec,MD_LOG_NAME);

	    /* New file to be opened */
	    if(md_log_fname[0] == '\0' || strcmp(tmp2, md_log_fname) != 0){
	      if(md_log_fid != NULL){
		fclose(md_log_fid); md_log_fid = NULL;
	      }

	      strcpy(md_log_fname,tmp2);
	      /* Add header if this is the first time we open the file
	       */
	      if(file_exists(md_log_fname)){
		md_log_fid = fopen(md_log_fname,"a");
	      } else { // New file, add header and reset MD event counter
		md_log_fid = fopen(md_log_fname,"a");
		fprintf(md_log_fid,"# year, month, day, hour, minute, second, trigger value, threshold, x-zones, y-zones, triggered zone, images saved, images saved before event, images saved after event, event number\n");
		md_trig_number = 0;
	      }
	    }

	    if(tmp != NULL){ free(tmp); }

	    tmp = make_time_string(current_frame.timestamp,current_frame.usec,MD_LOG_DATA);

	    /* Now the file should be open until program exits or the 
	       day changes */

	    int n_before = md_percentage * img_in_md_buffer / 100;

	    fprintf(md_log_fid,"%s,%.3lf,%d,%d,%d,%d,%d,%d,%d,%d\n",tmp,md_trigger_value,md_threshold,md_num_x_zones,md_num_y_zones,md_triggered_zone,img_in_md_buffer,n_before,img_in_md_buffer-n_before-1,md_trig_number);
	    fflush(md_log_fid);
	    free(tmp);
	  }
	  md_detection = md_i;
	  md_images_since_detection = 0;
	  break;
	}
      }
    }
  }

  if(md_detection != -1){
    md_images_since_detection++;
  }
  
  // Increment index where to store image
  md_i++;
  // If we are using the last slot of the image buffer, start from the zero
  if(md_i >= img_in_md_buffer){
    md_i = 0;
    md_buffer_ready = 1;
  }

  md_process_on = 0;

  pthread_exit(NULL);

}

void *md_save_images(void* ptr){

  char *fname = NULL;
  fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  int i = 0;

  /* Disable motion detection until the ring memory is being saved */
  md_saving_images = 1;

  /* Save ring memory to images on disc */
  if(!md_test){
    printf("Saving ring memory..\n");
  }
  for(i=0; i<img_in_md_buffer; i++){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname,md_frames[i].timestamp, md_frames[i].usec, TRIG);
    if(!md_test){
      if(verbose){
	printf("[%d] %s saved from ring memory\n",i,fname);
      }
      writeImage(fname, md_frames[i].width, md_frames[i].height, md_frames[i].img_buffer, "Finsprite motion detection at work");
    } else {
      if(verbose){
	printf("[%d] %s read from ring memory\n",i,fname);
      }
    }
  }

  md_trig_number++;

  free(fname); fname = NULL;

  if(!md_test){
    printf("%d images from ring memory saved.\n",img_in_md_buffer);
  }

  md_images_since_detection = -1;
  md_detection = -1;
  for(i=0; i<md_num_x_zones*md_num_y_zones; i++){
    md_zone_sums[i] = INT_MAX;
    md_zone_sums_new[i] = INT_MAX;
  }

  /* Start to fill the motion detection ring memory from the beginning */
  md_i = 0;
  md_buffer_ready = 0;

  /* Enable motion detection */
  md_saving_images = 0;

  pthread_exit(NULL);

}

void *process_stack(void* ptr){

  int i = 0;
  unsigned long min = ULONG_MAX;

  if(use_max || use_dif){
    if(peak_hold->img_buffer == NULL){
      init_peak_hold();
    }
  }
  if(use_latest){
    init_latest_img();
  }

  if(use_min || use_dif){
    if(min_hold->img_buffer == NULL){
      init_min_hold();
    }
  }

  if(use_ave){
    if(ave_stack->img_buffer == NULL){
      ave_stack->img_buffer = calloc(current_frame.width*current_frame.height, sizeof(*(ave_stack->img_buffer)));
      ave_stack->width = current_frame.width;
      ave_stack->height = current_frame.height;
      ave_stack->timestamp = current_frame.timestamp;
      ave_stack->usec = current_frame.usec;
      ave_stack->max_val = 0;
      ave_stack->min_val = ULONG_MAX;
      stack_timestamp = current_frame.timestamp;
      stack_usec = current_frame.usec;
    }
  }

  for(i=0;i<current_frame.width*current_frame.height;i++){

    if(use_max || use_dif){
      if(current_frame.img_buffer[i] > peak_hold->img_buffer[i]){
	peak_hold->img_buffer[i] = current_frame.img_buffer[i];
      }
    }

    if(use_latest){
      if(current_frame.img_buffer[i] > latest_img->img_buffer[i]){
	latest_img->img_buffer[i] = current_frame.img_buffer[i];
      }
    }

    if(use_min || use_dif){
      if(current_frame.img_buffer[i] < min_hold->img_buffer[i]){
	min_hold->img_buffer[i] = current_frame.img_buffer[i];
      }
    }

    if(use_ave){
      ave_stack->img_buffer[i] += current_frame.img_buffer[i];

      if(ave_stack->img_buffer[i] > ave_stack->max_val){
	ave_stack->max_val = ave_stack->img_buffer[i];
      }
      
      if(ave_stack->img_buffer[i] < min){
	min = ave_stack->img_buffer[i];
      }
    }
  }

  if(use_ave){
    ave_stack->num_imgs++;

    if(min > ave_stack->min_val){
      ave_stack->min_val = min;
    }
  }

  pthread_exit(NULL);

}

static int process_frame(const void * p, time_t timestamp, long usec){

  static int packed_value;

  int Y0, Y1, Y2, Y3;

  int x=0, y=0;
  
  int i = 0, j = 0;

  int maximum = 0;

  current_frame.width = fmt.fmt.pix.width;
  current_frame.height = fmt.fmt.pix.height;
  current_frame.timestamp = timestamp;
  current_frame.usec = usec;

  if(current_frame.img_buffer == NULL){
    current_frame.img_buffer = (unsigned char*)malloc(current_frame.width*current_frame.height*sizeof(unsigned char*));
  }

  while(j+3 < (fmt.fmt.pix.width * fmt.fmt.pix.height)){
    packed_value = *((int*)p+i);
    
    switch(fmt.fmt.pix.pixelformat){
    case V4L2_PIX_FMT_GREY:
      Y0 = (char)(packed_value & 0xFF);
      Y2 = (char)((packed_value >> 4) & 0xFF);
      Y1 = (char)((packed_value >> 8) & 0xFF);
      Y3 = (char)((packed_value >> 12) & 0xFF);

      break;
    case V4L2_PIX_FMT_UYVY:
    case V4L2_PIX_FMT_YUYV:

      Y0 = (char)(packed_value & 0xFF);
      Y2 = (char)((packed_value >> 8) & 0xFF);
      Y1 = (char)((packed_value >> 16) & 0xFF);
      Y3 = (char)((packed_value >> 24) & 0xFF);

      break;
    default:
      break;
    }

    // Strip sign values after shift (i.e. unsigned shift)
    Y0 = Y0 & 0xFF;
    Y1 = Y1 & 0xFF;
    Y2 = Y2 & 0xFF;
    Y3 = Y3 & 0xFF;

    x = (i+1) % (fmt.fmt.pix.width/2) - 1;
    y = (i+1) / (fmt.fmt.pix.width/2);
    j = x*2 + y*fmt.fmt.pix.width;

    if(j+3 > fmt.fmt.pix.width * fmt.fmt.pix.height){ break; }

    current_frame.img_buffer[j+0] = Y0;
    current_frame.img_buffer[j+1] = Y1;
    current_frame.img_buffer[j+2] = Y2;
    current_frame.img_buffer[j+3] = Y3;

    if(Y0+Y1+Y2+Y3 > maximum){
      maximum = Y0+Y1+Y2+Y3;
    }
    
    i++;
  }

  return(maximum);

}

void *save_stacks(void *ptr){

  saving_stacks = 1;

  /* Save stack information to log */
  if(save_log){
    if(stack_log_fname == NULL){
      stack_log_fname = (char*)malloc(FNAME_SIZE*sizeof(char));
      memset(stack_log_fname,'\0',FNAME_SIZE);
    }

    char tmp[FNAME_SIZE] = {'\0'};
    char *tmp2 = NULL;

    make_fname(tmp,stack_timestamp,stack_usec,STACK_LOG_NAME);

    /* New file to be opened */
    if(stack_log_fname[0] == '\0' || strcmp(tmp, stack_log_fname) != 0){
      if(stack_log_fid != NULL){
	fclose(stack_log_fid); stack_log_fid = NULL;
      }

      strcpy(stack_log_fname,tmp);
      if(file_exists(stack_log_fname)){
	stack_log_fid = fopen(stack_log_fname,"a");
      } else {
	stack_log_fid = fopen(stack_log_fname,"a");
	fprintf(stack_log_fid,"# start year, start month, start day, start hour, start minute, start second, end year, end month, end day, end hour, end minute, end second, stack length, images in stack, min stack saved, max stack saved, ave stack saved, dif stack saved\n");
      }
    }

    tmp2 = make_time_string(stack_timestamp,stack_usec,STACK_LOG_DATA);
    fprintf(stack_log_fid,"%s,%d,%d,%d,%d,%d,%d\n",tmp2,stack_length,num_img_in_save_stack,use_min,use_max,use_ave,use_dif);
    fflush(stack_log_fid);

    free(tmp2);

  }

  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);  

  if(use_ave){
    save_ave_stack();
  }
  
  if(use_dif){
    save_dif_stack();
  }

  if(use_max){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname, peak_hold_save->timestamp, peak_hold_save->usec, PEAK_HOLD);
    printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
    //stretch(peak_hold->img_buffer, peak_hold->width*peak_hold->height, 0.01, 0.99);
    writeImage(fname, peak_hold_save->width, peak_hold_save->height, peak_hold_save->img_buffer, "Finsprite peak-hold stack");
  }
  
  if(use_max || use_dif){
    free(peak_hold_save->img_buffer); peak_hold_save->img_buffer = NULL;
  }
  
  if(use_min){
    memset(fname,'\0',FNAME_SIZE);
    make_fname(fname, min_hold_save->timestamp, min_hold_save->usec, MIN_HOLD);
    printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
    //stretch(min_hold->img_buffer, min_hold->width*min_hold->height, 0.01, 0.99);
    writeImage(fname, min_hold_save->width, min_hold_save->height, min_hold_save->img_buffer, "Finsprite minimum stack");

    free(min_hold_save->img_buffer); min_hold_save->img_buffer = NULL;
  }
  
  if(use_min || use_dif){
    free(min_hold_save->img_buffer); min_hold_save->img_buffer = NULL;
  }
  
  num_img_in_save_stack = 0;
  //  stack_timestamp = current_frame.timestamp;

  saving_stacks = 0;

  free(fname);

  return NULL;

}

void *save_latest(void *ptr){

  saving_latest = 1;

  writeImage(latest_fname, latest_img_save->width, latest_img_save->height, latest_img_save->img_buffer, "Finsprite latest stack");
  
  free(latest_img_save->img_buffer); latest_img_save->img_buffer = NULL;
  
  saving_latest = 0;

  return NULL;

}

static void errno_exit(const char *s){
  fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
  exit(EXIT_FAILURE);
}

static int xioctl(int fd, int request, void *arg){
  int r;
  
  do {
    r = ioctl(fd, request, arg);
  } while(r == -1 && EINTR == errno);
  
  return r;
}

char *make_time_string(time_t timestamp, long usec, int type){

  char *time_string = (char*)malloc(100*sizeof(char));
  char *tmp = NULL;

  if(type < MD_LOG_NAME){
    strftime(time_string,30,"%Y-%m-%d_%H%M%S",gmtime(&(timestamp)));
  }

  switch(type){
  case TRIG:
    tmp = (char*)malloc(50*sizeof(char));
    sprintf(tmp,".%.6ld",usec);
    strcat(time_string,tmp);
    free(tmp);
    break;
  case MD_LOG_NAME:
  case STACK_LOG_NAME:
    strftime(time_string,30,"%Y-%m-%d",gmtime(&(timestamp)));
    break;
  case MD_LOG_DATA:
  case STACK_LOG_DATA:
    tmp = (char*)malloc(30*sizeof(char));
    strftime(time_string,30,"%Y,%m,%d,%H,%M,%S",gmtime(&(timestamp)));
    sprintf(tmp,".%.6ld",usec);
    strcat(time_string,tmp);
    if(type == STACK_LOG_DATA){
      memset(tmp,'\0',30);
      strftime(tmp,30,",%Y,%m,%d,%H,%M,%S",gmtime(&(stack_end_timestamp)));
      strcat(time_string,tmp);
      memset(tmp,'\0',30);
      sprintf(tmp,".%.6ld",stack_end_usec);
      strcat(time_string,tmp);
    }
    free(tmp);
    break;
  }
  return time_string;
}

void make_fname(char *fname, time_t timestamp, long usec, int type){
  char tmp[50] = {'\0'};
  mode_t dir_mode = S_IRWXU |S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;

  if(out_dir != NULL){ strcat(fname,out_dir); strcat(fname,"/"); }

  if(type == TRIG){
    sprintf(tmp,"%d",md_trig_year);
  } else {
    strftime(tmp,50,"%Y",gmtime(&(timestamp)));
  }
  strcat(fname,tmp);
  memset(tmp,'\0',50);

  if(mkdir(fname, dir_mode) < 0 && errno != EEXIST){
    printf("Unable to create %s\nSaving to current direcory: %d\n",fname,errno);
    // Empty path
    memset(fname,'\0',FNAME_SIZE);
  } else {

    if(type == TRIG){
      sprintf(tmp,"/%.2d",md_trig_month);
    } else {
      strftime(tmp,50,"/%m",gmtime(&(timestamp)));
    }
    strcat(fname,tmp); CLEAR(tmp);
    mkdir(fname, dir_mode);
    // Save only images to daily subfolders
    if(type < MD_LOG_NAME){
      if(type == TRIG){
	sprintf(tmp,"/%.2d",md_trig_day);
      } else {
	strftime(tmp,50,"/%d",gmtime(&(timestamp)));
      }
      strcat(fname,tmp); CLEAR(tmp);
      mkdir(fname, dir_mode);
    }

    /* Save motion detected images to event-folders, stacks to respective
       folders
    */
    switch(type){
    case MIN_HOLD:
      strcat(fname,"/min");
      break;
    case PEAK_HOLD:
      strcat(fname,"/max");
      break;
    case AVE:
      strcat(fname,"/ave");
      break;
    case DIF:
      strcat(fname,"/dif");
      break;
    case TRIG:
      sprintf(tmp,"/MD%.4d",md_trig_number);
      strcat(fname,tmp);
      break;
    }    

    mkdir(fname, dir_mode);
    memset(tmp,'\0',50);

    strcat(fname,"/");

  }

  /* If we want to have the daily image directory, skip filename creation */
  if(type != DAY_DIR){

    if(img_prefix != NULL){ strcat(fname,img_prefix); strcat(fname,"_"); }
    char *tmp2 = NULL;
    tmp2 = make_time_string(timestamp,usec,type);
    strcat(fname,tmp2); 
    free(tmp2);
    
    switch(type){
    case AVE:
      sprintf(tmp,"_UTC_%d_s_ave.png", stack_length);
      break;
    case PEAK_HOLD:
      sprintf(tmp,"_UTC_%d_s_max.png", stack_length);
      break;
    case MIN_HOLD:
      sprintf(tmp,"_UTC_%d_s_min.png", stack_length);
      break;
    case DIF:
      sprintf(tmp,"_UTC_%d_s_dif.png", stack_length);
      break;
    case TRIG:
      sprintf(tmp,"_UTC_md.png");
      break;
    case MD_LOG_NAME:
      sprintf(tmp,"_md.log");
      break;
    case STACK_LOG_NAME:
      sprintf(tmp,"_stacks.log");
      break;
    default:
      sprintf(tmp,".%ld_UTC.png",usec);
      break;
    }
    strcat(fname,tmp);
  }

}

void save_ave_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);

  unsigned char *out_frame = calloc(ave_stack_save->width*ave_stack_save->height,sizeof(unsigned char));

  if(ave_stack_save->img_buffer == NULL)
    printf("\n\nave_stack->img_buffer == NULL\n\n");

  for(i=0;i<ave_stack_save->width*ave_stack_save->height;i++){
    out_frame[i] = ave_stack_save->img_buffer[i]/ave_stack_save->num_imgs;
    //out_frame[i] = 255*ave_stack_save->img_buffer[i]/ave_stack_save->max_val;
    //out_frame[i] = 255*(ave_stack_save->img_buffer[i]-ave_stack_save->min_val)/(ave_stack_save->max_val-ave_stack_save->min_val);
  }

  make_fname(fname, ave_stack_save->timestamp, ave_stack_save->usec, AVE);
  //stretch(out_frame, ave_stack_save->width*ave_stack_save->height, 0.02, 0.98);
  printf("Saving %s, %d frames used.\n",fname,ave_stack_save->num_imgs);
  writeImage(fname, ave_stack_save->width, ave_stack_save->height, out_frame, "Finsprite average stack");
  
  free(out_frame);
  free(fname);

  free(ave_stack_save->img_buffer);
  ave_stack_save->img_buffer = NULL;
  ave_stack_save->min_val = ULONG_MAX;
  ave_stack_save->max_val = 0;
  ave_stack_save->num_imgs = 0;
}

void save_dif_stack(){

  int i;
  char *fname = (char*)malloc(FNAME_SIZE*sizeof(char));
  memset(fname,'\0',FNAME_SIZE);

  unsigned char *dif = calloc(min_hold_save->width*min_hold_save->height,sizeof(float));

  unsigned char *out_frame = calloc(min_hold_save->width*min_hold_save->height,sizeof(unsigned char));

  unsigned char dmin = UCHAR_MAX, dmax = 0;
  for(i=0;i<min_hold_save->width*min_hold_save->height;i++){
    dif[i] = peak_hold_save->img_buffer[i] - min_hold_save->img_buffer[i];
    if(dif[i] < dmin){ dmin = dif[i]; }
    if(dif[i] > dmax){ dmax = dif[i]; }
  }

  for(i=0;i<min_hold_save->width*min_hold_save->height;i++){
    out_frame[i] = 255*(dif[i]-dmin)/dmax;
  }
  
  free(dif); dif = NULL;

  make_fname(fname, min_hold_save->timestamp, min_hold_save->usec, DIF);
  printf("Saving %s, %d frames used.\n",fname,num_img_in_save_stack);
  //stretch(out_frame, min_hold->width*min_hold->height, 0.02, 0.98);
  writeImage(fname, min_hold_save->width, min_hold_save->height, out_frame, "Finsprite difference stack");
  
  free(out_frame); out_frame = NULL;
  free(fname); fname = NULL;

}

static int read_frame(void){
  struct v4l2_buffer buf;

  time_t frametime;
  
  int status = 0;
  void *status2 = NULL;

  CLEAR(buf);

  buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  buf.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
    switch(errno) {
    case EAGAIN:
      printf("EAGAIN\n");
      return 0;
      
    case EIO:
      /* Could ignore EIO, see spec. */
      printf("EIO\n");
      /* fall through */
      
    default:
      printf("default\n");
      errno_exit("VIDIOC_DQBUF");
    }
  }
    
  assert(buf.index < n_mmap_buffers);
  
  frametime = buf.timestamp.tv_sec;

  if(frametime > 0){
  //  if(!first_frame){
    if(process_frame(mmap_buffers[buf.index].start, frametime, buf.timestamp.tv_usec) > 0){

      num_img_in_stack++;

      //      pthread_t stack_thread, trig_thread;
      pthread_attr_t thread_attr;

      int stack_thread_used = 0, trig_thread_used = 0;

      pthread_attr_init(&thread_attr);
      pthread_attr_setdetachstate(&thread_attr, PTHREAD_CREATE_JOINABLE);

      /* If ring memory is ready to be saved, create a new thread for it */
      if(use_trig && !md_saving_images && 
	 100*md_images_since_detection/img_in_md_buffer >= md_percentage){
	md_saving_images = 1;
	status = pthread_create(&trig_save_thread, NULL, md_save_images, NULL);
	if(status){
	  printf("Thread \"md_save_images()\" failed with code %d\n",status);
	}
      }

      if(use_trig && !md_saving_images){
	trig_thread_used = 1;
	status = pthread_create(&trig_thread, &thread_attr, md_process, NULL);
	//status = pthread_create(&md_trig_thread, NULL, md_process, NULL);
	if(status){
	  printf("Thread \"md_process()\" failed with code %d\n",status);
	}
      }

      if(use_max || use_min || use_ave || use_dif){

	stack_thread_used = 1;
	status = pthread_create(&stack_thread, &thread_attr, process_stack, NULL);
	//status = pthread_create(&stack_thread, NULL, process_stack, NULL);
	if(status){
	  printf("Thread \"process_stack()\" failed with code %d\n",status);
	}
      }

      /* Join threads */
      
      pthread_attr_destroy(&thread_attr);
      if(stack_thread_used){
	status = pthread_join(stack_thread, &status2);
	if(status) {
	  printf("ERROR; return code from pthread_join() for stack_thread is %d\n", status);
	}
      }

      if(trig_thread_used){
	status = pthread_join(trig_thread, &status2);
	if(status) {
	  printf("ERROR; return code from pthread_join() for stack_thread is %d\n", status);
	}
      }
      
      if((use_ave || use_min || use_max || use_dif) && stack_length < difftime(current_frame.timestamp,stack_timestamp)){

	stack_end_timestamp = current_frame.timestamp;
	stack_end_usec = current_frame.usec;

	num_img_in_save_stack = num_img_in_stack;
	num_img_in_stack = 0;

	/* Swap stack pointers and create thread for stack saving */
	if(min_hold == &min_hold1){
	  min_hold = &min_hold2;
	  min_hold_save = &min_hold1;
	} else {
	  min_hold = &min_hold1;
	  min_hold_save = &min_hold2;
	}
	if(peak_hold == &peak_hold1){
	  peak_hold = &peak_hold2;
	  peak_hold_save = &peak_hold1;
	} else {
	  peak_hold = &peak_hold1;
	  peak_hold_save = &peak_hold2;
	}
	if(ave_stack == &ave_stack1){
	  ave_stack = &ave_stack2;
	  ave_stack_save = &ave_stack1;
	} else {
	  ave_stack = &ave_stack1;
	  ave_stack_save = &ave_stack2;
	}

	status = pthread_create(&stack_save_thread, NULL, save_stacks, NULL);
	if(status){
	  printf("Thread \"save_stacks()\" failed with code %d\n",status);
	}

      }

      if(use_latest && latest_length < difftime(current_frame.timestamp,latest_img_timestamp)){

	if(latest_img == &latest_img1){
	  latest_img = &latest_img2;
	  latest_img_save = &latest_img1;
	} else {
	  latest_img = &latest_img1;
	  latest_img_save = &latest_img2;
	}

	status = pthread_create(&latest_save_thread, NULL, save_latest, NULL);
	if(status){
	  printf("Thread \"save_latest()\" failed with code %d\n",status);
	}

      }









    }

  } else {
    first_frame = 0;
  }

  if(capture_length != 0){
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }

  return 1;
}

static void mainloop(void){
  time_t timenow;

  time(&start_time);
  time(&timenow);

  while(difftime(timenow,start_time) < capture_length) {

    for(;;) {
      fd_set fds;
      struct timeval tv;
      int r;
      
      FD_ZERO(&fds);
      FD_SET(fd, &fds);
      
      /* Timeout. */
      tv.tv_sec = 2;
      tv.tv_usec = 0;
      
      r = select(fd + 1, &fds, NULL, NULL, &tv);
      
      if(r == -1) {
	if(EINTR == errno)
	  continue;
	
	errno_exit("select");
      }
      
      if(r == 0) {
	fprintf(stderr, "select timeout\n");
	exit(EXIT_FAILURE);
      }
      
      if(read_frame())
	break;
      
      /* EAGAIN - continue select loop. */
    }
    time(&timenow);
  }

}

static void stop_capturing(void){
  enum v4l2_buf_type type;
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMOFF, &type) == -1)
    errno_exit("VIDIOC_STREAMOFF");

}

static void start_capturing(void){
  unsigned int i;
  enum v4l2_buf_type type;
  
  for(i = 0; i < n_mmap_buffers; ++i) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = i;
    
    if(xioctl(fd, VIDIOC_QBUF, &buf) == -1)
      errno_exit("VIDIOC_QBUF");
  }
  
  type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_STREAMON, &type) == -1){
    errno_exit("VIDIOC_STREAMON");
  }
}

static void uninit_device(void){
  unsigned int i;
  
  for(i = 0; i < n_mmap_buffers; ++i)
    if(munmap(mmap_buffers[i].start, mmap_buffers[i].length) == -1)
      errno_exit("munmap");
  
  free(mmap_buffers);
}

static void init_mmap(int n){
  struct v4l2_requestbuffers req;
  
  CLEAR(req);
  
  req.count = 2;
  req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  req.memory = V4L2_MEMORY_MMAP;
  
  if(xioctl(fd, VIDIOC_REQBUFS, &req) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s does not support memory mapping\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_REQBUFS");
    }
  }
  
  if(req.count < 2) {
    fprintf(stderr, "Insufficient buffer memory on %s\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  mmap_buffers = calloc(req.count, sizeof(*mmap_buffers));
  
  if(!mmap_buffers) {
    fprintf(stderr, "Out of memory\n");
    exit(EXIT_FAILURE);
  }
  
  for(n_mmap_buffers = 0; n_mmap_buffers < req.count; ++n_mmap_buffers) {
    struct v4l2_buffer buf;
    
    CLEAR(buf);
    
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    buf.index = n_mmap_buffers;
    
    if(xioctl(fd, VIDIOC_QUERYBUF, &buf) == -1)
      errno_exit("VIDIOC_QUERYBUF");
    
    mmap_buffers[n_mmap_buffers].length = buf.length;
    mmap_buffers[n_mmap_buffers].start =
      mmap(NULL /* start anywhere */,
	   buf.length,
	   PROT_READ | PROT_WRITE /* required */,
	   MAP_SHARED /* recommended */,
	   fd, buf.m.offset);
    
    if(MAP_FAILED == mmap_buffers[n_mmap_buffers].start)
      errno_exit("mmap");
  }
}

static void init_device(int img_in_mmap_buffer){
  struct v4l2_capability cap;
  struct v4l2_cropcap cropcap;
  struct v4l2_crop crop;
  unsigned int min;
  
  if(xioctl(fd, VIDIOC_QUERYCAP, &cap) == -1) {
    if(EINVAL == errno) {
      fprintf(stderr, "%s is no V4L2 device\n", dev_name);
      exit(EXIT_FAILURE);
    } else {
      errno_exit("VIDIOC_QUERYCAP");
    }
  }
  
  if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
    fprintf(stderr, "%s is no video capture device\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  if(!(cap.capabilities & V4L2_CAP_STREAMING)) {
    fprintf(stderr, "%s does not support streaming i/o\n",
	    dev_name);
    exit(EXIT_FAILURE);
  }
  
  /* Select video input, video standard and tune here. */
  
  CLEAR(cropcap);

  cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  
  if(xioctl(fd, VIDIOC_CROPCAP, &cropcap) == 0) {
    crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    crop.c = cropcap.defrect; /* reset to default */
    
    if(xioctl(fd, VIDIOC_S_CROP, &crop) == -1) {
      switch(errno) {
      case EINVAL:
	/* Cropping not supported. */
	break;
      default:
	/* Errors ignored. */
	break;
      }
    }
  } else {
    /* Errors ignored. */
  }
    
  CLEAR(fmt);
  
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = 720;
  fmt.fmt.pix.height = 576;
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

  if(xioctl(fd, VIDIOC_S_FMT, &fmt) == -1)
    errno_exit("VIDIOC_S_FMT");
  
  /* Note VIDIOC_S_FMT may change width and height. */
  
  /* Buggy driver paranoia. */
  min = fmt.fmt.pix.width * 2;
  if(fmt.fmt.pix.bytesperline < min)
    fmt.fmt.pix.bytesperline = min;
  min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
  if(fmt.fmt.pix.sizeimage < min)
    fmt.fmt.pix.sizeimage = min;
  
  init_mmap(img_in_mmap_buffer);

}

static void close_device(void){
  if(close(fd) == -1)
    errno_exit("close");
  
  fd = -1;
}

static void open_device(void){
  struct stat st; 
  
  if(stat(dev_name, &st) == -1) {
    fprintf(stderr, "Cannot identify '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
  
  if(!S_ISCHR(st.st_mode)) {
    fprintf(stderr, "%s is no device\n", dev_name);
    exit(EXIT_FAILURE);
  }
  
  fd = open(dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
  
  if(fd == -1) {
    fprintf(stderr, "Cannot open '%s': %d, %s\n",
	    dev_name, errno, strerror(errno));
    exit(EXIT_FAILURE);
  }
}

static void usage(FILE *fp, int argc, char **argv){
  fprintf(fp,
	  "Usage: %s [options]\n\n"
	  "Options:\n"
	  "-c | --capture-device <name> Video device name [/dev/video0]\n"
	  "-h | --help  Print this message\n"
	  "-n | --buffer <n>  Number of images in ring buffer [25]\n"
	  "                   Enables motion detection\n"
	  "-s | --stack-length <s>  Length of stacked period in seconds [60]\n"
	  "-m | --min  Enable minimum stacking [default: off]\n"
	  "-M | --max  Enable maximum stacking [default: off]\n"
	  "-a | --ave  Enable average stacking [default: off]\n"
	  "-d | --dif  Enable difference stacking [default: off]\n"
	  "-t | --trig  Enable motion detection with default ring buffer [default: off]\n"
	  "-e | --latest  Save \"latest.png\", an independent peak-hold stack\n"
	  "-T | --trigger-threshold  Threshold for motion detection [default: 1000]\n"
	  "-x | --x-regions  Number of motion detecion zones in horizontal direction [12]\n"
	  "-y | --y-regions  Number of motion detecion zones in vertical direction [9]\n"
	  "-k | --mask  Mask filename used to exclude areas from motion detect\n"
	  "-G | --trigger-test  Test how sensitive triggering is. No images are saved\n"
	  "-l | --capture-length <time>  Length of imaging period in seconds [unlimited]\n"
	  "-L | --latest-length <time>  Length of imaging period for \"latest.png\" in seconds [60]\n"
	  "-p | --prefix <prefix>  Image filename prefix [default: none]\n"
	  "-o | --out-dir <dir>  Image output directory [default: current directory]\n"
	  "                      The images are saved to <dir>/yyyy/mm/dd/\n"
	  "-O | --latest-fname <fname>  Filename with full path for saving \"latest.png\" [latest.png]\n"
	  "-N | --no-log  Turn of logging\n"
	  "\n",
	  argv[0]);
}

int main(int argc, char **argv){
  dev_name = "/dev/video0";

  for(;;) {
    int index;
    int c;
    
    c = getopt_long(argc, argv,
		    short_options, long_options,
		    &index);
    
    if(c == -1)
      break;
    
    switch(c) {
    case 0: /* getopt_long() flag */
      break;
    case 'c':
      dev_name = optarg;
      break;      
    case 'h':
      usage(stdout, argc, argv);
      exit(EXIT_SUCCESS);
    case 'n':
      sscanf(optarg,"%d",&img_in_md_buffer);
      use_trig = 1;
      break;
    case 's':  /* length of stacked period in secs */
      sscanf(optarg,"%d",&stack_length);
      break;
    case 'm': /* Enable minimum stack */
      use_min = 1;
      break;
    case 'M': /* Enable maximum stack */
      use_max = 1;
      break;
    case 'a': /* Enable average stack */
      use_ave = 1;
      break;
    case 't': /* Enable triggering (motion detection)*/
      use_trig = 1;
      break;
    case 'e': /* Enable saving of "latest.png" */
      use_latest = 1;
      break;
    case 'T': /* Trigger/motion detect threshold*/
      printf("%s\n",optarg);
      sscanf(optarg,"%d",&md_threshold);
      break;
    case 'x': /* Number of triggering regions in x-direction */
      sscanf(optarg,"%d",&md_num_x_zones);
      break;
    case 'y': /* Number of triggering regions in y-direction */
      sscanf(optarg,"%d",&md_num_y_zones);
      break;
    case 'k': /* Mask that can be used to exclude pixels from motion detection */
      md_mask_fname = optarg;
      break;
    case 'G':
      md_test = 1;
      use_trig = 1;
      break;
    case 'P':
      sscanf(optarg,"%d",&md_percentage);
      break;
    case 'd': /* Enable difference stack */
      use_dif = 1;
      break;
    case 'l': /* Length of imaging period */
      sscanf(optarg,"%ld",&capture_length);
      break;
    case 'p': /* Image prefix */
      img_prefix = optarg;
      break;
    case 'o': /* Output directory prefix */
      out_dir = optarg;
      break;
    case 'O': /* Filename for "latest.png" */
      latest_fname = optarg;
      use_latest = 1;
      break;
    case 'N': /* Disable logging */
      save_log = 0;
      break;
    case 'V': /* Verbose */
      verbose = 1;
      break;
    default:
      usage(stderr, argc, argv);
      exit(EXIT_FAILURE);
    }
  }

  if(!use_min && !use_max && !use_ave && !use_trig && !use_dif && !use_latest){
    printf("\nNothing to do, terminating.\n");
    printf("Specify at least one imaging mode!\n\n");
    usage(stderr, argc, argv);
    exit(EXIT_FAILURE);
  }

  if(md_test){
    use_min = 0;
    use_max = 0;
    use_ave = 0;
    use_dif = 0;
    use_latest = 0;
  }

  if(use_latest && latest_fname == NULL){
    latest_fname = (char*)malloc(10*sizeof(char));
    strcpy(latest_fname,latest_fname_default);
  }

  int i;
  ave_stack1.img_buffer = NULL;
  ave_stack1.min_val = ULONG_MAX;
  ave_stack1.max_val = 0;

  ave_stack2.img_buffer = NULL;
  ave_stack2.min_val = ULONG_MAX;
  ave_stack2.max_val = 0;

  ave_stack = &ave_stack1;
  ave_stack_save = NULL;

  peak_hold1.img_buffer = NULL;
  peak_hold2.img_buffer = NULL;
  peak_hold = &peak_hold1;
  peak_hold_save = NULL;

  min_hold1.img_buffer = NULL;
  min_hold2.img_buffer = NULL;
  min_hold = &min_hold1;
  min_hold_save = NULL;

  latest_img1.img_buffer = NULL;
  latest_img2.img_buffer = NULL;
  latest_img = &latest_img1;
  latest_img_save = NULL;

  if(use_trig){
    md_frames = calloc(img_in_md_buffer,sizeof(*md_frames));
    for(i=0; i<img_in_md_buffer; i++){
      md_frames[i].img_buffer = NULL;
      md_frames[i].width = -1;
      md_frames[i].height = -1;
    }
  }

  open_device();

  init_device(img_in_mmap_buffer);

  start_capturing();

  mainloop();

  /* Wait until all threads have finished */
  while(md_process_on || md_saving_images || saving_stacks){
  }

  if(capture_length != 0){
    stop_capturing();

    uninit_device();

    close_device();
  }

  if(use_trig){
    for(i=0; i < img_in_md_buffer; i++){
      if(md_frames[i].img_buffer != NULL){
	free(md_frames[i].img_buffer);
	md_frames[i].img_buffer = NULL;
      }
    }
    if(md_zone_sums != NULL){ free(md_zone_sums); }
    if(md_zone_sums_new != NULL){ free(md_zone_sums_new); }
    if(md_zone_num_pix != NULL){ free(md_zone_num_pix); }
    if(md_zone_reference != NULL){ free(md_zone_reference); }
  }

  if(md_frames != NULL){ free(md_frames); }
  if(peak_hold1.img_buffer != NULL){ free(peak_hold1.img_buffer); }
  if(peak_hold2.img_buffer != NULL){ free(peak_hold2.img_buffer); }
  if(min_hold1.img_buffer != NULL){ free(min_hold1.img_buffer); }
  if(min_hold2.img_buffer != NULL){ free(min_hold2.img_buffer); }
  if(ave_stack1.img_buffer != NULL){ free(ave_stack1.img_buffer); }
  if(ave_stack2.img_buffer != NULL){ free(ave_stack2.img_buffer); }
  if(latest_img1.img_buffer != NULL){ free(latest_img1.img_buffer); }
  if(latest_img2.img_buffer != NULL){ free(latest_img2.img_buffer); }

  if(current_frame.img_buffer != NULL){ free(current_frame.img_buffer); }

  if(md_log_fid != NULL){ fclose(md_log_fid); }
  if(stack_log_fid != NULL){ fclose(stack_log_fid); }

  exit(EXIT_SUCCESS);

  return 0;
}
